//#define CDP_CUDA

#include <math.h>

#include <cstdio> // for debugging

#include "stdafx.h"

#include "Model.h"

#define WANT_STREAM                  // include.h will get stream fns

#define WANT_MATH                    // include.h will get math fns

#include "newmatap.h"                // need matrix applications

#include "newmatio.h"                // need matrix output routines

#include "MersenneTwister.h"

#include "specialfunctions2.h"

#include "cdpprior.h"

#include "cdpbase.h"

#include "cdpresult.h"

#include "cdpresultem.h"

#include "cdp.h"

#include "cdp_em.h"





#if defined(CDP_TBB)

	#include "cdpt.h"

#endif



#if defined(CDP_CUDA)

	#include "CDPBaseCUDA.h"

	#include <cutil_inline.h>

	#include <cuda_runtime_api.h>

#endif

//int DIM,MEAN_CHD_DIM,PACK_DIM,CHD_DIM,LOGDET_OFFSET,DATA_PADDED_DIM,NCHUNKSIZE;



CDP_EM::CDP_EM()

{


}



CDP_EM::~CDP_EM(void)

{

	/*for (int i = 0; i < prior.N; i++) {

		delete [] mX[i];

	}

	delete [] mX;*/

}



bool CDP_EM::iterateEM(CDPResultEM& result)

{

	

	result.postLL = 0.0;

	

//	bool doPAlpha = true;
//	bool doMu = true;
//	bool doSigma = true;


	

	
	
#ifdef CDP_CUDA
	std::cout << "Updating Pi" << endl;
	cuda.updatePi(result.p[0], result.mu, result.Sigma_T_i, result.Sigma_log_det, result.sumiPi.Store(),&result.postLL);
		
	//std::cout << result.sumiPi;
	std::cout << "Updating Xbar" << endl;
	cuda.updateXbar(result.xbar, result.sumiPi);
	//std::cout << result.xbar;
	
#else
	getPi_ij(result);
	
	getXbar(result);
#endif
	
	updatePandAlpha(result);
	updateMu(result);



#ifdef CDP_CUDA
	std::cout << "Updating Sigma" << endl;
	cuda.updateSigma(result.Sigma, result.sumiPi, prior.Phi0, result.xbar, prior.gamma, prior.nu);
#else
		updateSigma(result);
#endif
			
//	std::cout << result.Sigma[0];
	
	
	LowerTriangularMatrix temp;
	SymmetricMatrix sigmaInv;
	Matrix workMat;
	
	for (int t=0; t<prior.T; t++) {
	
		temp << Cholesky(result.Sigma[t]);
					 
		result.Sigma_log_det[t] = msf.logdet(temp);
		result.Sigma_T_i[t] = temp.t().i();
					 
		//Some of the prior part of the posterior log-likelihood. I think this is OK.
		result.postLL -= (prior.nu + (double)result.D + 2.0)*( result.Sigma_log_det[t] / 2.0 );
					 
					 
		sigmaInv << result.Sigma_T_i[t] * result.Sigma_T_i[t].t();
					 
		workMat = prior.Phi0 * ( sigmaInv );
					 
		result.postLL -= 0.5 * workMat.Trace();
		result.postLL -= (1.0/(2.0 * prior.gamma)) * (( result.mu[t] - prior.m0 ) * sigmaInv * ( result.mu[t].t() - prior.m0.t() )).AsScalar();

	}	
	
	std::cout << "likelihood " << result.postLL << endl;

	

	//For testing
	
	//ofstream pi_ijfile;
	//pi_ijfile.open("pi_ij.txt");
	//pi_ijfile << setw(15) << setprecision(10) << result.pi_ij;


	/*ofstream xbarfile;
	xbarfile.open("xbartest.txt");
	ofstream sumiPifile;
	sumiPifile.open("sumiPi.txt");

	


	xbarfile << setw(10) << setprecision(5) << result.xbar;
	sumiPifile << setw(15) << setprecision(10) << result.sumiPi;
	std::cout << "Alpha = " << result.alpha[0] << endl;
	std::cout << "Does this make any sense? " << setprecision(15) << result.postLL << endl;*/

	
	return true;

}



// Test this more carefully. Most evidence points here ....

void CDP_EM::getPi_ij(CDPResultEM& result)	

{

	int i,t,j;

	
//	RowVector maxPdf(result.N);
	RowVector x_row(result.D);
	RowVector pi_row(result.T);
//	RowVector mixPdf(result.N);
//	Matrix pdf(result.N, result.T);
	
	LowerTriangularMatrix temp;

	double maxPdf;
	double mixPdf;

	double tmp = 0.0;

	for (i=0; i<result.N; i++) {

		for (t=0; t<result.T; t++) {

			//Get Normal Pdf Heights at data points
			//for (j =0; j < result.D; j++) {
			//	x_row[j] = mX[i][j];
			//}
			
			temp = (result.Sigma_T_i[t]).t();

			//result.pi_ij[i][t] = log(result.p[0][t]) + msf.mvnormpdf(mX[i], result.mu[t].Store(), result.Sigma_T_i[t], result.D, 1, result.Sigma_log_det[t]);
			result.pi_ij[i][t] = result.p[0][t] * msf.mvnormpdf(mX[i], result.mu[t].Store(), temp, result.D, 0, result.Sigma_log_det[t]);
			//pdf[i][t] =  msf.mvnormpdf(mX[i], result.mu[t].Store(), result.Sigma_T_i[t], result.D, 0, result.Sigma_log_det[t]);

		}

		/*maxPdf = result.pi_ij.Row(i+1).Maximum();

		
		for (t=0; t<result.T; t++) {
			result.pi_ij[i][t] = exp(result.pi_ij[i][t] - maxPdf);

		}*/

		

		mixPdf = result.pi_ij.Row(i+1).Sum();

		//std::cout << mixPdf << endl;

		// Add the data-likelihood to the posterior likelihood.

		//result.postLL += log(mixPdf) + maxPdf;
		result.postLL += log(mixPdf);



		// Normalize Pi_ij to sum to 1

		for (t=0; t<result.T; t++) {

			result.pi_ij[i][t] = result.pi_ij[i][t] / mixPdf;

		}

	}

	
	//std::cout << setw(25) << setprecision(20) <<  pdf;

	//std::cout << "data log-likelihood " << result.postLL << endl;
		

	// Get sumiPi

	for (t=1; t<=result.T; t++) {

		result.sumiPi(t) = result.pi_ij.Column(t).Sum();
		if (result.sumiPi(t) == 0.0) {
			result.sumiPi(t) = 0.0000000000000000000000000000001;
		}

	}



}



// This is just a standard matrix multiplication of pi_ij' * X but the way x is stored doesn't let me use newmat
// Looks correct.

void CDP_EM::getXbar(CDPResultEM& result){

	int t,d,n;

	double sum = 0.0;

	

	for (t=0; t<result.T; t++) {

		for (d=0; d<result.D; d++) {

			sum = 0.0;

			for (n=0; n<result.N; n++) {

				sum += result.pi_ij[n][t] * mX[n][d];

			}

			result.xbar[t][d] = sum / result.sumiPi[t];

		}

	}

		

	

}



// This updates the component weights along with V and sumiPi (needed for other computations)

// Looks correct.

void CDP_EM::updatePandAlpha(CDPResultEM& result)

{

	int t, r;

	

	double vepsilon = 0.000000000000001;

	double ww = 1.0;

	double sum = 0.0;

	

	RowVector partialSum(result.T);

	



	// Update V and Alpha (Newton's Method)

	partialSum[0] = (double)result.N;

	for (t=1; t<result.T; t++) {

		partialSum[t] = partialSum[t-1] - result.sumiPi[t-1];

	}

	

	for (r=0; r<10; r++) {

		// V

		for (t=0; t<result.T-1; t++) {

			result.pV[0][t] = result.sumiPi[t] / ( partialSum[t] + result.alpha[0] - 1.0);
			result.pV[0][t] = max(vepsilon, result.pV[0][t]);
			if (result.pV[0][t] >= 1.0) { result.pV[0][t] = 1.0 - vepsilon; }

		}
		result.pV[0][result.T-1] = 1.0;

		// Alpha

		sum = 0.0;
		for (t=0; t<result.T-1; t++) {sum += log( 1.0 - result.pV[0][t] );}
		result.alpha[0] = ( (double)result.T + prior.ee - 2.0) / (prior.ff - sum);

	}

	

	// Update component weights:	

	for (t=0; t<result.T-1; t++) {
		result.p[0][t] = ww * result.pV[0][t];
		ww = ww * (1.0 - result.pV[0][t]);

	}

	result.p[0][result.T-1] = ww;

	

	// Some of the posterior log-likelihood

	result.postLL += ( (double)result.T + prior.ee - 2.0 )*log(result.alpha[0]) - prior.ff*result.alpha[0];

	result.postLL += (result.alpha[0] - 1.0)*sum;

	//std::cout << ( (double)result.T + prior.ee - 2.0 )*log(result.alpha[0]) - prior.ff*result.alpha[0] + (result.alpha[0] - 1.0)*sum << endl;

}



void CDP_EM::updateMu(CDPResultEM& result){

	

	int t,d;

	RowVector xbarRow(result.D);

	for (t=0; t<result.T; t++) {

		for (d=0; d<result.D; d++) {

			xbarRow[d] = result.xbar[t][d];

		}

		

		result.mu[t] = ( ( prior.m0 / prior.gamma )  + xbarRow*result.sumiPi[t] ) * ( 1.0 / ( 1.0/prior.gamma + result.sumiPi[t]) ) ;

	}

	

}

extern "C" {
	
	void CPUcalcSigma(double* h_Sigma,int j, int JJ, int nn, int dd, double** h_X, double* h_xbar, double* h_pi_ij, double* h_sumipi,double* h_phi, double gamma, double nu){
		
		double sum=0.0;
		double xbarp,xbarq;
		int i,p,q;
			
			for(p=0;p<dd;p++){
				for(q=p;q<dd;q++){
					sum=0.0;
					xbarp = h_xbar[j*dd + p];
					xbarq = h_xbar[j*dd + q];	
					//printf("here");fflush(stdout);
					for(i=0;i<nn;i++){
						sum+=h_pi_ij[i*JJ + j] * (h_X[i][p] - xbarp) * (h_X[i][q] - xbarq);
					}
					//if(j==0){printf("%f\n",sum);}
					h_Sigma[p*dd+q] = ( h_phi[p*dd+q] + sum + (( h_sumipi[j] ) / (1.0 + gamma * h_sumipi[j] )) * xbarp*xbarq ) / ( h_sumipi[j] + 2.0 + 8.0 + (double)dd);
					
					h_Sigma[q*dd + p] = h_Sigma[p*dd+q];
				}
			}
		
		
		
	}
	
}



void CDP_EM::updateSigma(CDPResultEM& result)

{

	

	int n,t, d;
	RowVector xbarRow(result.D);
	RowVector xRow(result.D);
	RowVector diff(result.D);

	//SymmetricMatrix newSigma(result.D);
	Matrix newSigma(result.D,result.D); Matrix tmpPhi(result.D,result.D);
	SymmetricMatrix tmpMat(result.D);
	SymmetricMatrix sigmaInv(result.D);


	Matrix workMat(result.D,result.D);
	RowVector workVec(result.D);


	LowerTriangularMatrix temp;
		

	for (t=0; t<prior.T; t++) {

		
		/*newSigma = prior.Phi0;

		for (d=0; d<result.D; d++) {
			xbarRow[d] = result.xbar[t][d];
		}

		for (n=0; n<result.N; n++) {
			for (d=0; d<result.D; d++) {
				xRow[d] = mX[n][d];
			}
			diff = xRow - xbarRow;
			tmpMat <<  diff.t() * diff;
			newSigma += tmpMat * result.pi_ij[n][t] ;

		}

		diff = xbarRow - prior.m0;
		tmpMat << diff.t() * diff;
		newSigma += tmpMat * (( result.sumiPi[t] ) / (1.0 + prior.gamma * result.sumiPi[t] )) ;
		result.Sigma[t] =  newSigma / ( result.sumiPi[t] + prior.nu + (double)result.D + 2.0 );*/
		
		
		tmpPhi << prior.Phi0;
		
		CPUcalcSigma(newSigma.Store(), t , prior.T, prior.N, prior.D, mX, result.xbar.Store(), result.pi_ij.Store(), result.sumiPi.Store(),tmpPhi.Store(), prior.gamma, prior.nu);

		result.Sigma[t] << newSigma;
		/*temp << Cholesky(result.Sigma[t]);

		result.Sigma_log_det[t] = msf.logdet(temp);
		result.Sigma_T_i[t] = temp.t().i();

		//Some of the prior part of the posterior log-likelihood. I think this is OK.
		result.postLL -= (prior.nu + (double)result.D + 2.0)*( result.Sigma_log_det[t] / 2.0 );

	
		sigmaInv << result.Sigma_T_i[t] * result.Sigma_T_i[t].t();
		
		workMat = prior.Phi0 * ( sigmaInv );

		//result.postLL -= 0.5 * workMat.Trace();
		result.postLL -= (1.0/(2.0 * prior.gamma)) * (( result.mu[t] - prior.m0 ) * sigmaInv * ( result.mu[t].t() - prior.m0.t() )).AsScalar();*/

		
	}

	

}

void CDP_EM::EMAlgorithm(Model& model, MTRand& mt){
	
	
	
	double oldLL = 10000000000.0;
	
	double errTol;
	
	cdpfunctions.LoadData(model);
	mX = cdpfunctions.mX;

	
	prior.Init(model);
	cdpfunctions.prior = prior;
	
	CDPResultEM result(prior.J,prior.T,prior.N,prior.D);

	
#if defined(CDP_CUDA)
	
	if (model.mnGPU_Sample_Chunk_Size < 0) {
		
		NCHUNKSIZE = model.mnN;
		
	} else {
		
		NCHUNKSIZE = model.mnGPU_Sample_Chunk_Size;
		
	}
	
	cuda.initializeInstanceEM(model.startDevice,prior.J,prior.T, prior.N, 
								  
								  prior.D,model.numberDevices); 
	
	cuda.initializeDataEM(mX);
	
#endif
	
	cdpfunctions.SimulateFromPrior2(result,mt);
	

	
	
	
	// if any values are to be loaded from file, load them here
	
	cdpfunctions.LoadInits(model,result, mt);
	
	
	
	// see if we're dealing with a special case of J==1
	
	cdpfunctions.CheckSpecialCases(model,result);
	
	
	
#if defined(CDP_CUDA)
	
	unsigned int hTimer;
	
	cutilCheckError(cutCreateTimer(&hTimer));
	
	cutilCheckError(cutResetTimer(hTimer));
	
    cutilCheckError(cutStartTimer(hTimer));
	
#else
	
	time_t tStart, tEnd;
	
	//long tStart, tEnd;
	
	//tStart = clock();
	
	time(&tStart);
	
#endif
	
	
	
	// main EM loop

	
	errTol = log(1.0 + model.mnErrorPerTol);
	
	for (int it = 0; (it < model.mnIter) & (fabs(result.postLL - oldLL) > errTol) ; it++) {
		
		oldLL = result.postLL;
		
		
		
		std::cout << "it = " << (it+1) << endl;
		
		iterateEM(result);
		
		
		//result.SaveDraws();
		
		//std::cout << "EM iterate" << endl;
		
		
		
	}
	
	
	
#if defined(CDP_CUDA)
	
	cutilCheckError(cutStopTimer(hTimer));
	
    printf("GPU Processing time: %f (ms) \n", cutGetTimerValue(hTimer));
	
#else
	
	//tEnd = clock();
	
	time(&tEnd);
	
    cout << "time lapsed:" << difftime(tEnd,tStart)  << "seconds"<< endl;
	
#endif
	
	
	
	
	
	// save final parameter values
	
	result.SaveFinal();
	
	
	
}

/*
int main(int argc,char* argv[]) {

	

//	double oldLL = 10000000000.0;

//	double errTol;
	
	bool isEM = true;



  Model model;
 
  CDP_EM cdp;
  

//  CDP cdp;

  MTRand mt;

  string pfile = "parameters.txt";

  if (argc > 1) {

    pfile = string(argv[1]);

    if (pfile == "-DEFAULT" || pfile == "-Default" || pfile == "-default") {

      model.Save("default.parameters.txt");

      std::cout << "Default Parameters are saved in 'default.parameters.txt'" << endl;

      exit(0);

    }

  }

  if (!model.Load(pfile)) {

    std::cout << "Loading parameters file failed" << endl;

    exit (1);

  }



  mt.seed(model.mnSeed);

	cdp.EMAlgorithm(model,mt);

/*  cdp.LoadData(model);

  

  cdp.prior.Init(model);

  cdp.InitMCMCSteps(model);



  CDPResultEM result(cdp.prior.J,cdp.prior.T,cdp.prior.N,cdp.prior.D);



#if defined(CDP_CUDA)

	if (model.mnGPU_Sample_Chunk_Size < 0) {

			NCHUNKSIZE = model.mnN;

	} else {

			NCHUNKSIZE = model.mnGPU_Sample_Chunk_Size;

	}

	cdp.cuda.initializeInstanceEM(model.startDevice,cdp.prior.J,cdp.prior.T, cdp.prior.N, 

					cdp.prior.D,model.numberDevices); 

	cdp.cuda.initializeDataEM(cdp.mX);

#endif



  cdp.SimulateFromPrior2(result,mt);



  // if any values are to be loaded from file, load them here

  cdp.LoadInits(model,result, mt);



  // see if we're dealing with a special case of J==1

  cdp.CheckSpecialCases(model,result);



  #if defined(CDP_CUDA)

	unsigned int hTimer;

	cutilCheckError(cutCreateTimer(&hTimer));

	cutilCheckError(cutResetTimer(hTimer));

    cutilCheckError(cutStartTimer(hTimer));

  #else

	time_t tStart, tEnd;

	//long tStart, tEnd;

	//tStart = clock();

	time(&tStart);

  #endif



  // main EM loop

	errTol = log(1.0 + model.mnErrorPerTol);

  for (int it = 0; (it < model.mnIter) & (fabs(result.postLL - oldLL) > errTol) ; it++) {

	  oldLL = result.postLL;

	  

    std::cout << "it = " << (it+1) << endl;

	cdp.iterate(result);


	  //result.SaveDraws();

	  //std::cout << "EM iterate" << endl;



  }

	

  #if defined(CDP_CUDA)

	cutilCheckError(cutStopTimer(hTimer));

    printf("GPU Processing time: %f (ms) \n", cutGetTimerValue(hTimer));

  #else

	//tEnd = clock();

  time(&tEnd);

    cout << "time lapsed:" << difftime(tEnd,tStart)  << "seconds"<< endl;

  #endif



  

  // save final parameter values

 // result.SaveFinal();



  std::cout << "Done" << endl;

  return 0;

}
*/


